#ifndef ORCCBINDINGSSTACK_H
#define ORCCBINDINGSSTACK_H

#include "OrcCBindings.h"
#include "llvm/ADT/Triple.h"
#include "llvm/ExecutionEngine/Orc/CompileOnDemandLayer.h"
#include "llvm/ExecutionEngine/Orc/CompileUtils.h"
#include "llvm/ExecutionEngine/Orc/ExecutionUtils.h"
#include "llvm/ExecutionEngine/Orc/IRCompileLayer.h"
#include "llvm/ExecutionEngine/Orc/ObjectLinkingLayer.h"
#include "llvm/IR/LLVMContext.h"

namespace llvm {

class OrcCBindingsStack {
public:

  typedef orc::TargetAddress (*CExternalSymbolResolverFn)(const char *Name,
                                                          void *Ctx);

  typedef orc::JITCompileCallbackManagerBase CompileCallbackMgr;
  typedef orc::ObjectLinkingLayer<> ObjLayerT;
  typedef orc::IRCompileLayer<ObjLayerT> CompileLayerT;
  typedef orc::CompileOnDemandLayer<CompileLayerT, CompileCallbackMgr> CODLayerT;
  typedef std::function<
            std::unique_ptr<CompileCallbackMgr>(CompileLayerT&,
                                                RuntimeDyld::MemoryManager&,
                                                LLVMContext&)>
    CallbackManagerBuilder;

  typedef CODLayerT::ModuleSetHandleT ModuleHandleT;

  static CallbackManagerBuilder createCallbackManagerBuilder(Triple T);

  OrcCBindingsStack(std::unique_ptr<TargetMachine> TM, LLVMContext &Context,
                    CallbackManagerBuilder &BuildCallbackMgr)
    : TM(std::move(TM)), DL(this->TM->createDataLayout()),
      ObjectLayer(),
      CompileLayer(ObjectLayer, orc::SimpleCompiler(*this->TM)),
      CCMgr(BuildCallbackMgr(CompileLayer, CCMgrMemMgr, Context)),
      CODLayer(CompileLayer, *CCMgr, false),
      CXXRuntimeOverrides([this](const std::string &S) { return mangle(S); }) {}

  ~OrcCBindingsStack() {
    // Run any destructors registered with __cxa_atexit.
    CXXRuntimeOverrides.runDestructors();
    // Run any IR destructors.
    for (auto &DtorRunner : IRStaticDestructorRunners)
      DtorRunner.runViaLayer(CODLayer);
  }

  template <typename PtrTy>
  static PtrTy fromTargetAddress(orc::TargetAddress Addr) {
    return reinterpret_cast<PtrTy>(static_cast<uintptr_t>(Addr));
  }

  std::shared_ptr<RuntimeDyld::SymbolResolver>
  createResolver(CExternalSymbolResolverFn ExternalResolver,
                 void *ExternalResolverCtx) {
    auto Resolver = orc::createLambdaResolver(
      [this, ExternalResolver, ExternalResolverCtx](const std::string &Name) {
        // Search order:
        // 1. JIT'd symbols.
        // 2. Runtime overrides.
        // 3. External resolver (if present).

        if (auto Sym = CODLayer.findSymbol(Name, true))
          return RuntimeDyld::SymbolInfo(Sym.getAddress(),
                                         Sym.getFlags());
        if (auto Sym = CXXRuntimeOverrides.searchOverrides(Name))
          return Sym;

        if (ExternalResolver)
          return RuntimeDyld::SymbolInfo(ExternalResolver(Name.c_str(),
                                                          ExternalResolverCtx),
                                         llvm::JITSymbolFlags::Exported);

        return RuntimeDyld::SymbolInfo(nullptr);
      },
      [](const std::string &Name) {
        return RuntimeDyld::SymbolInfo(nullptr);
      }
    );

    return std::shared_ptr<RuntimeDyld::SymbolResolver>(std::move(Resolver));
  }

  ModuleHandleT addModule(std::unique_ptr<Module> M,
                          CExternalSymbolResolverFn ExternalResolver,
                          void *ExternalResolverCtx) {

    // Attach a data-layout if one isn't already present.
    if (M->getDataLayout().isDefault())
      M->setDataLayout(DL);

    // Record the static constructors and destructors. We have to do this before
    // we hand over ownership of the module to the JIT.
    std::vector<std::string> CtorNames, DtorNames;
    for (auto Ctor : orc::getConstructors(*M))
      CtorNames.push_back(mangle(Ctor.Func->getName()));
    for (auto Dtor : orc::getDestructors(*M))
      DtorNames.push_back(mangle(Dtor.Func->getName()));

    // Create the resolver.
    auto Resolver = createResolver(ExternalResolver, ExternalResolverCtx);

    // Add the module to the JIT.
    std::vector<std::unique_ptr<Module>> S;
    S.push_back(std::move(M));

    auto H = CODLayer.addModuleSet(std::move(S), nullptr, std::move(Resolver));

    // Run the static constructors, and save the static destructor runner for
    // execution when the JIT is torn down.
    orc::CtorDtorRunner<CODLayerT> CtorRunner(std::move(CtorNames), H);
    CtorRunner.runViaLayer(CODLayer);

    IRStaticDestructorRunners.emplace_back(std::move(DtorNames), H);

    return H;
  }

  orc::JITSymbol findSymbol(const std::string &Name) {
    return CODLayer.findSymbol(mangle(Name), true);
  }

  orc::JITSymbol findSymbolIn(ModuleHandleT H, const std::string &Name) {
    return CODLayer.findSymbolIn(H, mangle(Name), true);
  }

private:

  std::string mangle(const std::string &Name) {
    std::string MangledName;
    {
      raw_string_ostream MangledNameStream(MangledName);
      Mangler::getNameWithPrefix(MangledNameStream, Name, DL);
    }
    return MangledName;
  }

  std::unique_ptr<TargetMachine> TM;
  DataLayout DL;
  SectionMemoryManager CCMgrMemMgr;

  ObjLayerT ObjectLayer;
  CompileLayerT CompileLayer;
  std::unique_ptr<CompileCallbackMgr> CCMgr;
  CODLayerT CODLayer;

  orc::LocalCXXRuntimeOverrides CXXRuntimeOverrides;
  std::vector<orc::CtorDtorRunner<CODLayerT>> IRStaticDestructorRunners;
};

DEFINE_SIMPLE_CONVERSION_FUNCTIONS(OrcCBindingsStack, OrcJITStackRef);

} // end namespace llvm

#endif
